@@ -1,13 +1,16 @@
-# Copyright (C) 2020  The Software Heritage developers
+# Copyright (C) 2020-2021  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
 import asyncio
 from dataclasses import dataclass, field
+import json
+import logging
+import os
 from pathlib import Path
-from typing import Any, AsyncIterator, List
-import urllib.parse
+import re
+from typing import Any, AsyncIterator, Dict, List, Optional, cast
 
 from swh.fuse.fs.entry import (
     EntryMode,
@@ -17,17 +20,16 @@ from swh.fuse.fs.entry import (
     FuseSymlinkEntry,
 )
 from swh.model.from_disk import DentryPerms
-from swh.model.identifiers import CONTENT, DIRECTORY, RELEASE, REVISION, SNAPSHOT, SWHID
+from swh.model.hashutil import hash_to_bytes, hash_to_hex
+from swh.model.identifiers import CoreSWHID, ObjectType
+
+SWHID_REGEXP = r"swh:1:(cnt|dir|rel|rev|snp):[0-9a-f]{40}"
 
 
 @dataclass
 class Content(FuseFileEntry):
     """ Software Heritage content artifact.
 
-    Attributes:
-        swhid: Software Heritage persistent identifier
-        prefetch: optional prefetched metadata used to set entry attributes
-
     Content leaves (AKA blobs) are represented on disks as regular files,
     containing the corresponding bytes, as archived.
 
@@ -36,8 +38,9 @@ class Content(FuseFileEntry):
     directory, the permissions of the `archive/SWHID` file will be arbitrary and
     not meaningful (e.g., `0x644`). """
 
-    swhid: SWHID
+    swhid: CoreSWHID
     prefetch: Any = None
+    """optional prefetched metadata used to set entry attributes"""
 
     async def get_content(self) -> bytes:
         data = await self.fuse.get_blob(self.swhid)
@@ -49,16 +52,13 @@ class Content(FuseFileEntry):
         if self.prefetch:
             return self.prefetch["length"]
         else:
-            return len(await self.get_content())
+            return await super().size()
 
 
 @dataclass
 class Directory(FuseDirEntry):
     """ Software Heritage directory artifact.
 
-    Attributes:
-        swhid: Software Heritage persistent identifier
-
     Directory nodes are represented as directories on the file-system,
     containing one entry for each entry of the archived directory. Entry names
     and other metadata, including permissions, will correspond to the archived
@@ -68,7 +68,7 @@ class Directory(FuseDirEntry):
     So it is possible that, in the context of a directory, a file is presented
     as writable, whereas actually writing to it will fail with `EPERM`. """
 
-    swhid: SWHID
+    swhid: CoreSWHID
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
         metadata = await self.fuse.get_metadata(self.swhid)
@@ -79,12 +79,24 @@ class Directory(FuseDirEntry):
                 # Archived permissions for directories are always set to
                 # 0o040000 so use a read-only permission instead
                 int(EntryMode.RDONLY_DIR)
-                if swhid.object_type == DIRECTORY
+                if swhid.object_type == ObjectType.DIRECTORY
                 else entry["perms"]
             )
 
-            # 1. Regular file
-            if swhid.object_type == CONTENT:
+            # 1. Symlink (check symlink first because condition is less restrictive)
+            if mode == DentryPerms.symlink:
+                target = b""
+                try:
+                    # Symlink target is stored in the blob content
+                    target = await self.fuse.get_blob(swhid)
+                except Exception:
+                    pass  # Ignore error and create a (broken) symlink anyway
+
+                yield self.create_child(
+                    FuseSymlinkEntry, name=name, target=target,
+                )
+            # 2. Regular file
+            elif swhid.object_type == ObjectType.CONTENT:
                 yield self.create_child(
                     Content,
                     name=name,
@@ -94,24 +106,20 @@ class Directory(FuseDirEntry):
                     # attributes without additional Software Heritage API call
                     prefetch=entry,
                 )
-            # 2. Regular directory
-            elif swhid.object_type == DIRECTORY:
+            # 3. Regular directory
+            elif swhid.object_type == ObjectType.DIRECTORY:
                 yield self.create_child(
                     Directory, name=name, mode=mode, swhid=swhid,
                 )
-            # 3. Symlink
-            elif mode == DentryPerms.symlink:
-                yield self.create_child(
-                    FuseSymlinkEntry,
-                    name=name,
-                    # Symlink target is stored in the blob content
-                    target=await self.fuse.get_blob(swhid),
-                )
             # 4. Submodule
-            elif swhid.object_type == REVISION:
-                # Make sure the revision metadata is fetched and create a
-                # symlink to distinguish it with regular directories
-                await self.fuse.get_metadata(swhid)
+            elif swhid.object_type == ObjectType.REVISION:
+                try:
+                    # Make sure the revision metadata is fetched and create a
+                    # symlink to distinguish it with regular directories
+                    await self.fuse.get_metadata(swhid)
+                except Exception:
+                    pass  # Ignore error and create a (broken) symlink anyway
+
                 yield self.create_child(
                     FuseSymlinkEntry,
                     name=name,
@@ -125,9 +133,6 @@ class Directory(FuseDirEntry):
 class Revision(FuseDirEntry):
     """ Software Heritage revision artifact.
 
-    Attributes:
-        swhid: Software Heritage persistent identifier
-
     Revision (AKA commit) nodes are represented on the file-system as
     directories with the following entries:
 
@@ -141,12 +146,12 @@ class Revision(FuseDirEntry):
       has at least one parent commit (which is the most common case). When
       present it is a symlink pointing into `parents/1/`
     - `history`: a virtual directory listing all its revision ancestors, sorted
-      in reverse topological order. Each entry is a symlink pointing into
-      `archive/SWHID`.
+      in reverse topological order. The history can be listed through
+      `by-date/`, `by-hash/` or `by-page/` with each its own sharding policy.
     - `meta.json`: metadata for the current node, as a symlink pointing to the
-      relevant `meta/<SWHID>.json` file """
+      relevant `archive/<SWHID>.json` file """
 
-    swhid: SWHID
+    swhid: CoreSWHID
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
         metadata = await self.fuse.get_metadata(self.swhid)
@@ -163,7 +168,7 @@ class Revision(FuseDirEntry):
         yield self.create_child(
             FuseSymlinkEntry,
             name="meta.json",
-            target=Path(root_path, f"meta/{self.swhid}.json"),
+            target=Path(root_path, f"archive/{self.swhid}.json"),
         )
         yield self.create_child(
             RevisionParents,
@@ -189,7 +194,7 @@ class Revision(FuseDirEntry):
 class RevisionParents(FuseDirEntry):
     """ Revision virtual `parents/` directory """
 
-    parents: List[SWHID]
+    parents: List[CoreSWHID]
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
         root_path = self.get_relative_root_path()
@@ -205,24 +210,41 @@ class RevisionParents(FuseDirEntry):
 class RevisionHistory(FuseDirEntry):
     """ Revision virtual `history/` directory """
 
-    swhid: SWHID
+    swhid: CoreSWHID
 
-    async def prefill_caches(self) -> None:
+    async def prefill_by_date_cache(self, by_date_dir: FuseDirEntry) -> None:
         history = await self.fuse.get_history(self.swhid)
+        nb_api_calls = 0
         for swhid in history:
+            cache = await self.fuse.cache.metadata.get(swhid)
+            if cache:
+                continue
+
             await self.fuse.get_metadata(swhid)
+            # The by-date/ directory is cached temporarily in direntry, and
+            # invalidated + updated every 100 API calls
+            nb_api_calls += 1
+            if nb_api_calls % 100 == 0:
+                self.fuse.cache.direntry.invalidate(by_date_dir)
+        # Make sure to have the latest entries once the prefilling is done
+        self.fuse.cache.direntry.invalidate(by_date_dir)
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
-        # Run it concurrently because of the many API calls necessary
-        asyncio.create_task(self.prefill_caches())
-
-        yield self.create_child(
+        by_date_dir = cast(
             RevisionHistoryShardByDate,
-            name="by-date",
-            mode=int(EntryMode.RDONLY_DIR),
-            history_swhid=self.swhid,
+            self.create_child(
+                RevisionHistoryShardByDate,
+                name="by-date",
+                mode=int(EntryMode.RDONLY_DIR),
+                history_swhid=self.swhid,
+            ),
         )
 
+        # Run it concurrently because of the many API calls necessary
+        asyncio.create_task(self.prefill_by_date_cache(by_date_dir))
+
+        yield by_date_dir
+
         yield self.create_child(
             RevisionHistoryShardByHash,
             name="by-hash",
@@ -242,11 +264,12 @@ class RevisionHistory(FuseDirEntry):
 class RevisionHistoryShardByDate(FuseDirEntry):
     """ Revision virtual `history/by-date` sharded directory """
 
-    history_swhid: SWHID
+    history_swhid: CoreSWHID
     prefix: str = field(default="")
     is_status_done: bool = field(default=False)
 
     DATE_FMT = "{year:04d}/{month:02d}/{day:02d}/"
+    ENTRIES_REGEXP = re.compile(r"^([0-9]{2,4})|(" + SWHID_REGEXP + ")$")
 
     @dataclass
     class StatusFile(FuseFileEntry):
@@ -254,32 +277,43 @@ class RevisionHistoryShardByDate(FuseDirEntry):
 
         name: str = field(init=False, default=".status")
         mode: int = field(init=False, default=int(EntryMode.RDONLY_FILE))
-        done: int
-        todo: int
+        history_swhid: CoreSWHID
+
+        def __post_init__(self):
+            super().__post_init__()
+            # This is the only case where we do not want the kernel to cache the file
+            self.file_info_attrs["keep_cache"] = False
+            self.file_info_attrs["direct_io"] = True
 
         async def get_content(self) -> bytes:
-            fmt = f"Done: {self.done}/{self.todo}\n"
+            history_full = await self.fuse.get_history(self.history_swhid)
+            history_cached = await self.fuse.cache.history.get_with_date_prefix(
+                self.history_swhid, date_prefix=""
+            )
+            fmt = f"Done: {len(history_cached)}/{len(history_full)}\n"
             return fmt.encode()
 
-        async def size(self) -> int:
-            return len(await self.get_content())
+    def __post_init__(self):
+        super().__post_init__()
+        # Create the status file only once so we can easily remove it when the
+        # entire history is fetched
+        self.status_file = self.create_child(
+            RevisionHistoryShardByDate.StatusFile, history_swhid=self.history_swhid
+        )
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
-        history = await self.fuse.get_history(self.history_swhid)
-        # Only check for cached revisions since fetching all of them with the
-        # Web API would take too long
-        swhids = await self.fuse.cache.metadata.get_cached_subset(history)
+        history_full = await self.fuse.get_history(self.history_swhid)
+        # Only check for cached revisions with the appropriate prefix, since
+        # fetching all of them with the Web API would take too long
+        history_cached = await self.fuse.cache.history.get_with_date_prefix(
+            self.history_swhid, date_prefix=self.prefix
+        )
 
         depth = self.prefix.count("/")
         root_path = self.get_relative_root_path()
         sharded_dirs = set()
 
-        for swhid in swhids:
-            meta = await self.fuse.cache.metadata.get(swhid)
-            date = meta["date"]
-            sharded_name = self.DATE_FMT.format(
-                year=date.year, month=date.month, day=date.day
-            )
+        for (swhid, sharded_name) in history_cached:
             if not sharded_name.startswith(self.prefix):
                 continue
 
@@ -302,23 +336,22 @@ class RevisionHistoryShardByDate(FuseDirEntry):
                         history_swhid=self.history_swhid,
                     )
 
-        self.is_status_done = len(swhids) == len(history)
-        if not self.is_status_done and depth == 0:
-            yield self.create_child(
-                RevisionHistoryShardByDate.StatusFile,
-                done=len(swhids),
-                todo=len(history),
-            )
+        self.is_status_done = len(history_cached) == len(history_full)
+        if self.is_status_done:
+            self.fuse._remove_inode(self.status_file.inode)
+        elif not self.is_status_done and depth == 0:
+            yield self.status_file
 
 
 @dataclass
 class RevisionHistoryShardByHash(FuseDirEntry):
     """ Revision virtual `history/by-hash` sharded directory """
 
-    history_swhid: SWHID
+    history_swhid: CoreSWHID
     prefix: str = field(default="")
 
     SHARDING_LENGTH = 2
+    ENTRIES_REGEXP = re.compile(r"^([a-f0-9]+)|(" + SWHID_REGEXP + ")$")
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
         history = await self.fuse.get_history(self.history_swhid)
@@ -326,7 +359,7 @@ class RevisionHistoryShardByHash(FuseDirEntry):
         if self.prefix:
             root_path = self.get_relative_root_path()
             for swhid in history:
-                if swhid.object_id.startswith(self.prefix):
+                if swhid.object_id.startswith(hash_to_bytes(self.prefix)):
                     yield self.create_child(
                         FuseSymlinkEntry,
                         name=str(swhid),
@@ -336,7 +369,7 @@ class RevisionHistoryShardByHash(FuseDirEntry):
         else:
             sharded_dirs = set()
             for swhid in history:
-                next_prefix = swhid.object_id[: self.SHARDING_LENGTH]
+                next_prefix = hash_to_hex(swhid.object_id)[: self.SHARDING_LENGTH]
                 if next_prefix not in sharded_dirs:
                     sharded_dirs.add(next_prefix)
                     yield self.create_child(
@@ -352,11 +385,12 @@ class RevisionHistoryShardByHash(FuseDirEntry):
 class RevisionHistoryShardByPage(FuseDirEntry):
     """ Revision virtual `history/by-page` sharded directory """
 
-    history_swhid: SWHID
-    prefix: int = field(default=None)
+    history_swhid: CoreSWHID
+    prefix: Optional[int] = field(default=None)
 
     PAGE_SIZE = 10_000
     PAGE_FMT = "{page_number:03d}"
+    ENTRIES_REGEXP = re.compile(r"^([0-9]+)|(" + SWHID_REGEXP + ")$")
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
         history = await self.fuse.get_history(self.history_swhid)
@@ -389,9 +423,6 @@ class RevisionHistoryShardByPage(FuseDirEntry):
 class Release(FuseDirEntry):
     """ Software Heritage release artifact.
 
-    Attributes:
-        swhid: Software Heritage persistent identifier
-
     Release nodes are represented on the file-system as directories with the
     following entries:
 
@@ -401,18 +432,18 @@ class Release(FuseDirEntry):
       (transitively) resolves to a directory. When present it is a symlink
       pointing into `archive/` to the SWHID of the given directory
     - `meta.json`: metadata for the current node, as a symlink pointing to the
-      relevant `meta/<SWHID>.json` file """
+      relevant `archive/<SWHID>.json` file """
 
-    swhid: SWHID
+    swhid: CoreSWHID
 
-    async def find_root_directory(self, swhid: SWHID) -> SWHID:
-        if swhid.object_type == RELEASE:
+    async def find_root_directory(self, swhid: CoreSWHID) -> Optional[CoreSWHID]:
+        if swhid.object_type == ObjectType.RELEASE:
             metadata = await self.fuse.get_metadata(swhid)
             return await self.find_root_directory(metadata["target"])
-        elif swhid.object_type == REVISION:
+        elif swhid.object_type == ObjectType.REVISION:
             metadata = await self.fuse.get_metadata(swhid)
             return metadata["directory"]
-        elif swhid.object_type == DIRECTORY:
+        elif swhid.object_type == ObjectType.DIRECTORY:
             return swhid
         else:
             return None
@@ -424,7 +455,7 @@ class Release(FuseDirEntry):
         yield self.create_child(
             FuseSymlinkEntry,
             name="meta.json",
-            target=Path(root_path, f"meta/{self.swhid}.json"),
+            target=Path(root_path, f"archive/{self.swhid}.json"),
         )
 
         target = metadata["target"]
@@ -451,47 +482,140 @@ class Release(FuseDirEntry):
 class ReleaseType(FuseFileEntry):
     """ Release type virtual file """
 
-    target_type: str
+    target_type: ObjectType
 
     async def get_content(self) -> bytes:
-        return str.encode(self.target_type + "\n")
-
-    async def size(self) -> int:
-        return len(await self.get_content())
+        return str.encode(self.target_type.name.lower() + "\n")
 
 
 @dataclass
 class Snapshot(FuseDirEntry):
     """ Software Heritage snapshot artifact.
 
-    Attributes:
-        swhid: Software Heritage persistent identifier
+    Snapshot nodes are represented on the file-system as recursive directories
+    following the branch names structure. For example, a branch named
+    ``refs/tags/v1.0`` will be represented as a ``refs`` directory containing a
+    ``tags`` directory containing a ``v1.0`` symlink pointing to the branch
+    target SWHID. """
 
-    Snapshot nodes are represented on the file-system as directories with one
-    entry for each branch in the snapshot. Each entry is a symlink pointing into
-    `archive/` to the branch target SWHID. Branch names are URL encoded (hence
-    '/' are replaced with '%2F'). """
-
-    swhid: SWHID
+    swhid: CoreSWHID
+    prefix: str = field(default="")
 
     async def compute_entries(self) -> AsyncIterator[FuseEntry]:
         metadata = await self.fuse.get_metadata(self.swhid)
         root_path = self.get_relative_root_path()
 
+        subdirs = set()
         for branch_name, branch_meta in metadata.items():
-            # Mangle branch name to create a valid UNIX filename
-            name = urllib.parse.quote_plus(branch_name)
+            if not branch_name.startswith(self.prefix):
+                continue
+
+            next_subdirs = branch_name[len(self.prefix) :].split("/")
+            next_prefix = next_subdirs[0]
+
+            if len(next_subdirs) == 1:
+                # Non-alias targets are symlinks to their corresponding archived
+                # artifact, whereas alias targets are relative symlinks to the
+                # corresponding snapshot directory entry.
+                target_type = branch_meta["target_type"]
+                target_raw = branch_meta["target"]
+                if target_type == "alias":
+                    prefix = Path(branch_name).parent
+                    target = os.path.relpath(target_raw, prefix)
+                else:
+                    target = f"{root_path}/archive/{target_raw}"
+
+                yield self.create_child(
+                    FuseSymlinkEntry, name=next_prefix, target=Path(target),
+                )
+            else:
+                subdirs.add(next_prefix)
+
+        for subdir in subdirs:
             yield self.create_child(
-                FuseSymlinkEntry,
-                name=name,
-                target=Path(root_path, f"archive/{branch_meta['target']}"),
+                Snapshot,
+                name=subdir,
+                mode=int(EntryMode.RDONLY_DIR),
+                swhid=self.swhid,
+                prefix=f"{self.prefix}{subdir}/",
             )
 
 
+@dataclass
+class Origin(FuseDirEntry):
+    """ Software Heritage origin artifact.
+
+    Origin nodes are represented on the file-system as directories with one
+    entry for each origin visit.
+
+    The visits directories are named after the visit date (`YYYY-MM-DD`, if
+    multiple visits occur the same day only the first one is kept). Each visit
+    directory contains a `meta.json` with associated metadata for the origin
+    node, and potentially a `snapshot` symlink pointing to the visit's snapshot
+    node. """
+
+    DATE_FMT = "{year:04d}-{month:02d}-{day:02d}"
+    ENTRIES_REGEXP = re.compile(r"^[0-9]{4}-[0-9]{2}-[0-9]{2}$")
+
+    async def compute_entries(self) -> AsyncIterator[FuseEntry]:
+        # The origin's name is always its URL (encoded to create a valid UNIX filename)
+        visits = await self.fuse.get_visits(self.name)
+
+        seen_date = set()
+        for visit in visits:
+            date = visit["date"]
+            name = self.DATE_FMT.format(year=date.year, month=date.month, day=date.day)
+
+            if name in seen_date:
+                logging.debug(
+                    "Conflict date on origin: %s, %s", visit["origin"], str(name)
+                )
+            else:
+                seen_date.add(name)
+                yield self.create_child(
+                    OriginVisit, name=name, mode=int(EntryMode.RDONLY_DIR), meta=visit,
+                )
+
+
+@dataclass
+class OriginVisit(FuseDirEntry):
+    """ Origin visit virtual directory """
+
+    meta: Dict[str, Any]
+
+    @dataclass
+    class MetaFile(FuseFileEntry):
+        content: str
+
+        async def get_content(self) -> bytes:
+            return str.encode(self.content + "\n")
+
+    async def compute_entries(self) -> AsyncIterator[FuseEntry]:
+        snapshot_swhid = self.meta["snapshot"]
+        if snapshot_swhid:
+            root_path = self.get_relative_root_path()
+            yield self.create_child(
+                FuseSymlinkEntry,
+                name="snapshot",
+                target=Path(root_path, f"archive/{snapshot_swhid}"),
+            )
+
+        yield self.create_child(
+            OriginVisit.MetaFile,
+            name="meta.json",
+            mode=int(EntryMode.RDONLY_FILE),
+            content=json.dumps(
+                self.meta,
+                indent=self.fuse.conf["json-indent"],
+                default=lambda x: str(x),
+            ),
+        )
+
+
 OBJTYPE_GETTERS = {
-    CONTENT: Content,
-    DIRECTORY: Directory,
-    REVISION: Revision,
-    RELEASE: Release,
-    SNAPSHOT: Snapshot,
+    ObjectType.CONTENT: Content,
+    ObjectType.DIRECTORY: Directory,
+    ObjectType.REVISION: Revision,
+    ObjectType.RELEASE: Release,
+    ObjectType.SNAPSHOT: Snapshot,
 }
