@@ -6,42 +6,35 @@
 from enum import Enum
 from typing import Union
 
-from swh.model.identifiers import (
-    CONTENT,
-    DIRECTORY,
-    RELEASE,
-    REVISION,
-    SNAPSHOT,
-    SWHID,
-    parse_swhid,
-)
+from swh.model.hashutil import hash_to_hex
+from swh.model.identifiers import CoreSWHID, ObjectType
 
 GRAPH_API_REQUEST = Enum("GRAPH_API_REQUEST", "HISTORY")
 
 
-def swhid_to_web_url(swhid: Union[SWHID, str], raw: bool = False) -> str:
+def swhid_to_web_url(swhid: Union[CoreSWHID, str], raw: bool = False) -> str:
     if isinstance(swhid, str):
-        swhid = parse_swhid(swhid)
+        swhid = CoreSWHID.from_string(swhid)
 
     prefix = {
-        CONTENT: "content/sha1_git:",
-        DIRECTORY: "directory/",
-        REVISION: "revision/",
-        RELEASE: "release/",
-        SNAPSHOT: "snapshot/",
+        ObjectType.CONTENT: "content/sha1_git:",
+        ObjectType.DIRECTORY: "directory/",
+        ObjectType.REVISION: "revision/",
+        ObjectType.RELEASE: "release/",
+        ObjectType.SNAPSHOT: "snapshot/",
     }
 
-    url = f"{prefix[swhid.object_type]}{swhid.object_id}/"
+    url = f"{prefix[swhid.object_type]}{hash_to_hex(swhid.object_id)}/"
     if raw:
         url += "raw/"
     return url
 
 
 def swhid_to_graph_url(
-    swhid: Union[SWHID, str], request_type: GRAPH_API_REQUEST
+    swhid: Union[CoreSWHID, str], request_type: GRAPH_API_REQUEST
 ) -> str:
     if isinstance(swhid, str):
-        swhid = parse_swhid(swhid)
+        swhid = CoreSWHID.from_string(swhid)
 
     prefix = {
         GRAPH_API_REQUEST.HISTORY: "graph/visit/edges/",
